# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function

import collections
import fnmatch
import logging
import os.path
import re
import subprocess
import sys

from reno import utils

LOG = logging.getLogger(__name__)


def _get_current_version(reporoot, branch=None):
    """Return the current version of the repository.

    If the repo appears to contain a python project, use setup.py to
    get the version so pbr (if used) can do its thing. Otherwise, use
    git describe.

    """
    cmd = ['git', 'describe', '--tags']
    if branch is not None:
        cmd.append(branch)
    try:
        result = utils.check_output(cmd, cwd=reporoot).strip()
        if '-' in result:
            # Descriptions that come after a commit look like
            # 2.0.0-1-abcde, and we want to remove the SHA value from
            # the end since we only care about the version number
            # itself, but we need to recognize that the change is
            # unreleased so keep the -1 part.
            result, dash, ignore = result.rpartition('-')
    except subprocess.CalledProcessError:
        # This probably means there are no tags.
        result = '0.0.0'
    return result


def get_file_at_commit(reporoot, filename, sha):
    "Return the contents of the file if it exists at the commit, or None."
    try:
        return utils.check_output(
            ['git', 'show', '%s:%s' % (sha, filename)],
            cwd=reporoot,
        )
    except subprocess.CalledProcessError:
        return None


def _file_exists_at_commit(reporoot, filename, sha):
    "Return true if the file exists at the given commit."
    return bool(get_file_at_commit(reporoot, filename, sha))


def _get_unique_id(filename):
    base = os.path.basename(filename)
    root, ext = os.path.splitext(base)
    uniqueid = root[-16:]
    if '-' in uniqueid:
        # This is an older file with the UUID at the beginning
        # of the name.
        uniqueid = root[:16]
    return uniqueid


# The git log output from _get_tags_on_branch() looks like this sample
# from the openstack/nova repository for git 1.9.1:
#
# git log --simplify-by-decoration --pretty="%d"
# (HEAD, origin/master, origin/HEAD, gerrit/master, master)
# (apu/master)
# (tag: 13.0.0.0b1)
# (tag: 12.0.0.0rc3, tag: 12.0.0)
#
# And this for git 1.7.1 (RHEL):
#
# $ git log --simplify-by-decoration --pretty="%d"
# (HEAD, origin/master, origin/HEAD, master)
# (tag: 13.0.0.0b1)
# (tag: 12.0.0.0rc3, tag: 12.0.0)
# (tag: 12.0.0.0rc2)
# (tag: 2015.1.0rc3, tag: 2015.1.0)
# ...
# (tag: folsom-2)
# (tag: folsom-1)
# (essex-1)
# (diablo-2)
# (diablo-1)
# (2011.2)
#
# The difference in the tags with "tag:" and without appears to be
# caused by some being annotated and others not.
#
# So we might have multiple tags on a given commit, and we might have
# lines that have no tags or are completely blank, and we might have
# "tag:" or not. This pattern is used to find the tag entries on each
# line, ignoring tags that don't look like version numbers.
TAG_RE = re.compile('''
    (?:[(]|tag:\s)  # look for tag: prefix and drop
    ((?:[\d.ab]|rc)+)  # digits, a, b, and rc cover regular and pre-releases
    [,)]  # possible trailing comma or closing paren
''', flags=re.VERBOSE | re.UNICODE)
PRE_RELEASE_RE = re.compile('''
    \.(\d+(?:[ab]|rc)+\d*)$
''', flags=re.VERBOSE | re.UNICODE)


def _branch_or_eol_tag(reporoot, branch):
    "If a stable branch is deleted, convert the name to the eol tag."
    branch_results = utils.check_output(
        ['git', 'branch', '--list', '-a'],
        cwd=reporoot,
    )
    branches = []
    for line in branch_results.splitlines():
        # Clean up whitespace and the * marking the "current" branch.
        line = line.strip().lstrip('*').strip()
        if line.startswith('remotes/'):
            line = line[8:]
        branches.append(line)
    if branch in branches:
        return branch
    return branch.rpartition('/')[-1] + '-eol'


def _get_version_tags_on_branch(reporoot, branch):
    """Return tags from the branch, in date order.

    Need to get the list of tags in the right order, because the topo
    search breaks the date ordering. Use git to ask for the tags in
    order, rather than trying to sort them, because many repositories
    have "non-standard" tags or have renumbered projects (from
    date-based to SemVer), for which sorting would require complex
    logic.

    """
    tags = []
    tag_cmd = [
        'git', 'log',
        '--simplify-by-decoration',
        '--pretty="%d"',
    ]
    if branch:
        tag_cmd.append(_branch_or_eol_tag(reporoot, branch))
    LOG.debug('running %s' % ' '.join(tag_cmd))
    tag_results = utils.check_output(tag_cmd, cwd=reporoot)
    LOG.debug(tag_results)
    for line in tag_results.splitlines():
        LOG.debug('line %r' % line)
        for match in TAG_RE.findall(line):
            tags.append(match)
    return tags


def get_notes_by_version(reporoot, notesdir, branch=None,
                         collapse_pre_releases=True,
                         earliest_version=None):
    """Return an OrderedDict mapping versions to lists of notes files.

    The versions are presented in reverse chronological order.

    Notes files are associated with the earliest version for which
    they were available, regardless of whether they changed later.

    :param reporoot: Path to the root of the git repository.
    :type reporoot: str
    :param notesdir: The directory under *reporoot* with the release notes.
    :type notesdir: str
    :param branch: The name of the branch to scan. Defaults to current.
    :type branch: str
    :param collapse_pre_releases: When true, merge pre-release versions
        into the final release, if it is present.
    :type collapse_pre_releases: bool
    """

    LOG.debug('scanning %s/%s (branch=%s)' % (reporoot, notesdir, branch))

    # Determine all of the tags known on the branch, in their date
    # order. We scan the commit history in topological order to ensure
    # we have the commits in the right version, so we might encounter
    # the tags in a different order during that phase.
    versions_by_date = _get_version_tags_on_branch(reporoot, branch)
    LOG.debug('versions by date %r' % (versions_by_date,))
    versions = []
    earliest_seen = collections.OrderedDict()

    # Determine the current version, which might be an unreleased or
    # dev version if there are unreleased commits at the head of the
    # branch in question. Since the version may not already be known,
    # make sure it is in the list of versions by date. And since it is
    # the most recent version, go ahead and insert it at the front of
    # the list.
    current_version = _get_current_version(reporoot, branch)
    LOG.debug('current repository version: %s' % current_version)
    if current_version not in versions_by_date:
        LOG.debug('adding %s to versions by date' % current_version)
        versions_by_date.insert(0, current_version)

    # Remember the most current filename for each id, to allow for
    # renames.
    last_name_by_id = {}

    # Remember uniqueids that have had files deleted.
    uniqueids_deleted = collections.defaultdict(set)

    # FIXME(dhellmann): This might need to be more line-oriented for
    # longer histories.
    log_cmd = [
        'git', 'log',
        '--topo-order',  # force traversal order rather than date order
        '--pretty=%x00%H %d',  # output contents in parsable format
        '--name-only'  # only include the names of the files in the patch
    ]
    if branch is not None:
        log_cmd.append(_branch_or_eol_tag(reporoot, branch))
    LOG.debug('running %s' % ' '.join(log_cmd))
    history_results = utils.check_output(log_cmd, cwd=reporoot)
    history = history_results.split('\x00')
    current_version = current_version
    for h in history:
        h = h.strip()
        if not h:
            continue
        # print(h)

        hlines = h.splitlines()

        # The first line of the block will include the SHA and may
        # include tags, the other lines are filenames.
        sha = hlines[0].split(' ')[0]
        tags = TAG_RE.findall(hlines[0])
        # Filter the files based on the notes directory we were
        # given. We cannot do this in the git log command directly
        # because it means we end up skipping some of the tags if the
        # commits being tagged don't include any release note
        # files. Even if this list ends up empty, we continue doing
        # the other processing so that we record all of the known
        # versions.
        filenames = []
        for f in hlines[2:]:
            if fnmatch.fnmatch(f, notesdir + '/*.yaml'):
                filenames.append(f)
            elif fnmatch.fnmatch(f, notesdir + '/*'):
                LOG.warn('found and ignored extra file %s', f)

        # If there are no tags in this block, assume the most recently
        # seen version.
        if not tags:
            tags = [current_version]
        else:
            current_version = tags[0]
            LOG.debug('%s has tags %s (%r), updating current version to %s' %
                      (sha, tags, hlines[0], current_version))

        # Remember each version we have seen.
        if current_version not in versions:
            LOG.debug('%s is a new version' % current_version)
            versions.append(current_version)

        LOG.debug('%s contains files %s' % (sha, filenames))

        # Remember the files seen, using their UUID suffix as a unique id.
        for f in filenames:
            # Updated as older tags are found, handling edits to release
            # notes.
            uniqueid = _get_unique_id(f)
            LOG.debug('%s: found file %s',
                      uniqueid, f)
            LOG.debug('%s: setting earliest reference to %s' %
                      (uniqueid, tags[0]))
            earliest_seen[uniqueid] = tags[0]
            if uniqueid in last_name_by_id:
                # We already have a filename for this id from a
                # new commit, so use that one in case the name has
                # changed.
                LOG.debug('%s: was seen before in %s',
                          uniqueid, last_name_by_id[uniqueid])
                continue
            elif _file_exists_at_commit(reporoot, f, sha):
                LOG.debug('%s: looking for %s in deleted files %s',
                          uniqueid, f, uniqueids_deleted[uniqueid])
                if f in uniqueids_deleted[uniqueid]:
                    # The file exists in the commit, but was deleted
                    # later in the history.
                    LOG.debug('%s: skipping deleted file %s',
                              uniqueid, f)
                else:
                    # Remember this filename as the most recent version of
                    # the unique id we have seen, in case the name
                    # changed from an older commit.
                    last_name_by_id[uniqueid] = (f, sha)
                    LOG.debug('%s: remembering %s as filename',
                              uniqueid, f)
            else:
                # Track files that have been deleted. The rename logic
                # above checks for repeated references to files that
                # are deleted later, and the inversion logic below
                # checks for any remaining values and skips those
                # entries.
                LOG.debug('%s: saw a file that no longer exists',
                          uniqueid)
                uniqueids_deleted[uniqueid].add(f)
                LOG.debug('%s: deleted files %s',
                          uniqueid, uniqueids_deleted[uniqueid])

    # Invert earliest_seen to make a list of notes files for each
    # version.
    files_and_tags = collections.OrderedDict()
    for v in versions:
        files_and_tags[v] = []
    # Produce a list of the actual files present in the repository. If
    # a note is removed, this step should let us ignore it.
    for uniqueid, version in earliest_seen.items():
        try:
            base, sha = last_name_by_id[uniqueid]
            if base in uniqueids_deleted.get(uniqueid, set()):
                LOG.debug('skipping deleted note %s' % uniqueid)
                continue
            files_and_tags[version].append((base, sha))
        except KeyError:
            # Unable to find the file again, skip it to avoid breaking
            # the build.
            msg = ('[reno] unable to find file associated '
                   'with unique id %r, skipping') % uniqueid
            LOG.debug(msg)
            print(msg, file=sys.stderr)

    # Combine pre-releases into the final release, if we are told to
    # and the final release exists.
    if collapse_pre_releases:
        collapsing = files_and_tags
        files_and_tags = collections.OrderedDict()
        for ov in versions_by_date:
            if ov not in collapsing:
                # We don't need to collapse this one because there are
                # no notes attached to it.
                continue
            pre_release_match = PRE_RELEASE_RE.search(ov)
            LOG.debug('checking %r', ov)
            if pre_release_match:
                # Remove the trailing pre-release part of the version
                # from the string.
                pre_rel_str = pre_release_match.groups()[0]
                canonical_ver = ov[:-len(pre_rel_str)].rstrip('.')
                if canonical_ver not in versions_by_date:
                    # This canonical version was never tagged, so we
                    # do not want to collapse the pre-releases. Reset
                    # to the original version.
                    canonical_ver = ov
                else:
                    LOG.debug('combining into %r', canonical_ver)
            else:
                canonical_ver = ov
            if canonical_ver not in files_and_tags:
                files_and_tags[canonical_ver] = []
            files_and_tags[canonical_ver].extend(collapsing[ov])

    # Only return the parts of files_and_tags that actually have
    # filenames associated with the versions.
    trimmed = collections.OrderedDict()
    for ov in versions_by_date:
        if not files_and_tags.get(ov):
            continue
        # Sort the notes associated with the version so they are in a
        # deterministic order, to avoid having the same data result in
        # different output depending on random factors. Earlier
        # versions of the scanner assumed the notes were recorded in
        # chronological order based on the commit date, but with the
        # change to use topological sorting that is no longer
        # necessarily true. We want the notes to always show up in the
        # same order, but it doesn't really matter what order that is,
        # so just sort based on the unique id.
        trimmed[ov] = sorted(files_and_tags[ov])
        # If we have been told to stop at a version, we can do that
        # now.
        if earliest_version and ov == earliest_version:
            break

    LOG.debug('[reno] found %d versions and %d files',
              len(trimmed.keys()), sum(len(ov) for ov in trimmed.values()))
    return trimmed



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import logging
import sys

from reno import cache
from reno import create
from reno import defaults
from reno import lister
from reno import report

_query_args = [
    (('--version',),
     dict(default=[],
          action='append',
          help='the version(s) to include, defaults to all')),
    (('--branch',),
     dict(default=None,
          help='the branch to scan, defaults to the current')),
    (('--collapse-pre-releases',),
     dict(action='store_true',
          default=True,
          help='combine pre-releases with their final release')),
    (('--no-collapse-pre-releases',),
     dict(action='store_false',
          dest='collapse_pre_releases',
          help='show pre-releases separately')),
    (('--earliest-version',),
     dict(default=None,
          help='stop when this version is reached in the history')),
    (('--ignore-cache',),
     dict(default=False,
          action='store_true',
          help='if there is a cache file present, do not use it')),
]


def _build_query_arg_group(parser):
    group = parser.add_argument_group('query')
    for args, kwds in _query_args:
        group.add_argument(*args, **kwds)


def main(argv=sys.argv[1:]):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-v', '--verbose',
        dest='verbosity',
        default=logging.INFO,
        help='produce more output',
        action='store_const',
        const=logging.DEBUG,
    )
    parser.add_argument(
        '-q', '--quiet',
        dest='verbosity',
        action='store_const',
        const=logging.WARN,
        help='produce less output',
    )
    parser.add_argument(
        '--rel-notes-dir', '-d',
        dest='relnotesdir',
        default=defaults.RELEASE_NOTES_SUBDIR,
        help='location of release notes YAML files',
    )
    subparsers = parser.add_subparsers(
        title='commands',
    )

    do_new = subparsers.add_parser(
        'new',
        help='create a new note',
    )
    do_new.add_argument(
        'slug',
        help='descriptive title of note (keep it short)',
    )
    do_new.set_defaults(func=create.create_cmd)

    do_list = subparsers.add_parser(
        'list',
        help='list notes files based on query arguments',
    )
    _build_query_arg_group(do_list)
    do_list.add_argument(
        'reporoot',
        help='root of the git repository',
    )
    do_list.set_defaults(func=lister.list_cmd)

    do_report = subparsers.add_parser(
        'report',
        help='generate release notes report',
    )
    do_report.add_argument(
        'reporoot',
        help='root of the git repository',
    )
    do_report.add_argument(
        '--output', '-o',
        default=None,
        help='output filename, defaults to stdout',
    )
    _build_query_arg_group(do_report)
    do_report.set_defaults(func=report.report_cmd)

    do_cache = subparsers.add_parser(
        'cache',
        help='generate release notes cache',
    )
    do_cache.add_argument(
        'reporoot',
        help='root of the git repository',
    )
    do_cache.add_argument(
        '--output', '-o',
        default=None,
        help=('output filename, '
              'defaults to the cache file within the notesdir, '
              'use "-" for stdout'),
    )
    _build_query_arg_group(do_cache)
    do_cache.set_defaults(func=cache.cache_cmd)

    args = parser.parse_args()

    logging.basicConfig(
        level=args.verbosity,
        format='%(message)s',
    )

    return args.func(args)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function

import os

from reno import utils


_TEMPLATE = """\
---
prelude: >
    Replace this text with content to appear at the
    top of the section for this release. All of the
    prelude content is merged together and then rendered
    separately from the items listed in other parts of
    the file, so the text needs to be worded so that
    both the prelude and the other items make sense
    when read independently. This may mean repeating
    some details. Not every release note
    requires a prelude. Usually only notes describing
    major features or adding release theme details should
    have a prelude.
features:
  - List new features here, or remove this section.
    All of the list items in this section are combined
    when the release notes are rendered, so the text
    needs to be worded so that it does not depend on any
    information only available in another section, such
    as the prelude. This may mean repeating some details.
issues:
  - List known issues here, or remove this section.
    All of the list items in this section are combined
    when the release notes are rendered, so the text
    needs to be worded so that it does not depend on any
    information only available in another section, such
    as the prelude. This may mean repeating some details.
upgrade:
  - List upgrade notes here, or remove this section.
    All of the list items in this section are combined
    when the release notes are rendered, so the text
    needs to be worded so that it does not depend on any
    information only available in another section, such
    as the prelude. This may mean repeating some details.
deprecations:
  - List deprecations notes here, or remove this section.
    All of the list items in this section are combined
    when the release notes are rendered, so the text
    needs to be worded so that it does not depend on any
    information only available in another section, such
    as the prelude. This may mean repeating some details.
critical:
  - Add critical notes here, or remove this section.
    All of the list items in this section are combined
    when the release notes are rendered, so the text
    needs to be worded so that it does not depend on any
    information only available in another section, such
    as the prelude. This may mean repeating some details.
security:
  - Add security notes here, or remove this section.
    All of the list items in this section are combined
    when the release notes are rendered, so the text
    needs to be worded so that it does not depend on any
    information only available in another section, such
    as the prelude. This may mean repeating some details.
fixes:
  - Add normal bug fixes here, or remove this section.
    All of the list items in this section are combined
    when the release notes are rendered, so the text
    needs to be worded so that it does not depend on any
    information only available in another section, such
    as the prelude. This may mean repeating some details.
other:
  - Add other notes here, or remove this section.
    All of the list items in this section are combined
    when the release notes are rendered, so the text
    needs to be worded so that it does not depend on any
    information only available in another section, such
    as the prelude. This may mean repeating some details.
"""


def _pick_note_file_name(notesdir, slug):
    "Pick a unique name in notesdir."
    for i in range(50):
        newid = utils.get_random_string()
        notefilename = os.path.join(notesdir, '%s-%s.yaml' % (slug, newid))
        if not os.path.exists(notefilename):
            return notefilename
    else:
        raise ValueError(
            'Unable to generate unique random filename '
            'in %s after 50 tries' % notesdir,
        )


def _make_note_file(filename):
    notesdir = os.path.dirname(filename)
    if not os.path.exists(notesdir):
        os.makedirs(notesdir)
    with open(filename, 'w') as f:
        f.write(_TEMPLATE)


def create_cmd(args):
    "Create a new release note file from the template."
    notesdir = utils.get_notes_dir(args)
    # NOTE(dhellmann): There is a short race window where we might try
    # to pick a name that does not exist, then overwrite the file if
    # it is created before we try to write it. This isn't a problem
    # because this command is expected to be run by one developer in
    # their local git tree, and so there should not be any concurrency
    # concern.
    slug = args.slug.replace(' ', '-')
    filename = _pick_note_file_name(notesdir, slug)
    _make_note_file(filename)
    print('Created new notes file in %s' % filename)
    return



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools
import logging
import os.path
import re
import subprocess
import textwrap

from reno import create
from reno import scanner
from reno.tests import base
from reno import utils

import fixtures
import mock
from testtools.content import text_content


_SETUP_TEMPLATE = """
import setuptools
try:
    import multiprocessing  # noqa
except ImportError:
    pass

setuptools.setup(
    setup_requires=['pbr'],
    pbr=True)
"""

_CFG_TEMPLATE = """
[metadata]
name = testpkg
summary = Test Package

[files]
packages =
    testpkg
"""


class GPGKeyFixture(fixtures.Fixture):
    """Creates a GPG key for testing.

    It's recommended that this be used in concert with a unique home
    directory.
    """

    def setUp(self):
        super(GPGKeyFixture, self).setUp()
        tempdir = self.useFixture(fixtures.TempDir())
        gnupg_version_re = re.compile('^gpg\s.*\s([\d+])\.([\d+])\.([\d+])')
        gnupg_version = utils.check_output(['gpg', '--version'],
                                           cwd=tempdir.path)
        for line in gnupg_version.split('\n'):
            gnupg_version = gnupg_version_re.match(line)
            if gnupg_version:
                gnupg_version = (int(gnupg_version.group(1)),
                                 int(gnupg_version.group(2)),
                                 int(gnupg_version.group(3)))
                break
        else:
            if gnupg_version is None:
                gnupg_version = (0, 0, 0)
        config_file = tempdir.path + '/key-config'
        f = open(config_file, 'wt')
        try:
            if gnupg_version[0] == 2 and gnupg_version[1] >= 1:
                f.write("""
                %no-protection
                %transient-key
                """)
            f.write("""
            %no-ask-passphrase
            Key-Type: RSA
            Name-Real: Example Key
            Name-Comment: N/A
            Name-Email: example@example.com
            Expire-Date: 2d
            Preferences: (setpref)
            %commit
            """)
        finally:
            f.close()
        # Note that --quick-random (--debug-quick-random in GnuPG 2.x)
        # does not have a corresponding preferences file setting and
        # must be passed explicitly on the command line instead
        if gnupg_version[0] == 1:
            gnupg_random = '--quick-random'
        elif gnupg_version[0] >= 2:
            gnupg_random = '--debug-quick-random'
        else:
            gnupg_random = ''
        cmd = ['gpg', '--gen-key', '--batch']
        if gnupg_random:
            cmd.append(gnupg_random)
        cmd.append(config_file)
        subprocess.check_call(cmd, cwd=tempdir.path)


class Base(base.TestCase):

    def _run_git(self, *args):
        return utils.check_output(
            ['git'] + list(args),
            cwd=self.reporoot,
        )

    def _git_setup(self):
        os.makedirs(self.reporoot)
        self._run_git('init', '.')
        self._run_git('config', '--local', 'user.email', 'example@example.com')
        self._run_git('config', '--local', 'user.name', 'reno developer')
        self._run_git('config', '--local', 'user.signingkey',
                      'example@example.com')

    def _git_commit(self, message='commit message'):
        self._run_git('add', '.')
        self._run_git('commit', '-m', message)

    def _add_other_file(self, name):
        with open(os.path.join(self.reporoot, name), 'w') as f:
            f.write('adding %s\n' % name)
        self._git_commit('add %s' % name)

    def _add_notes_file(self, slug='slug', commit=True, legacy=False):
        n = self.get_note_num()
        if legacy:
            basename = '%016x-%s.yaml' % (n, slug)
        else:
            basename = '%s-%016x.yaml' % (slug, n)
        filename = os.path.join(self.reporoot, 'releasenotes', 'notes',
                                basename)
        create._make_note_file(filename)
        self._git_commit('add %s' % basename)
        return os.path.join('releasenotes', 'notes', basename)

    def _make_python_package(self):
        setup_name = os.path.join(self.reporoot, 'setup.py')
        with open(setup_name, 'w') as f:
            f.write(_SETUP_TEMPLATE)
        cfg_name = os.path.join(self.reporoot, 'setup.cfg')
        with open(cfg_name, 'w') as f:
            f.write(_CFG_TEMPLATE)
        pkgdir = os.path.join(self.reporoot, 'testpkg')
        os.makedirs(pkgdir)
        init = os.path.join(pkgdir, '__init__.py')
        with open(init, 'w') as f:
            f.write("Test package")
        self._git_commit('add test package')

    def setUp(self):
        super(Base, self).setUp()
        self.logger = self.useFixture(
            fixtures.FakeLogger(
                format='%(levelname)8s %(name)s %(message)s',
                level=logging.DEBUG,
                nuke_handlers=True,
            )
        )
        # Older git does not have config --local, so create a temporary home
        # directory to permit using git config --global without stepping on
        # developer configuration.
        self.useFixture(fixtures.TempHomeDir())
        self.useFixture(GPGKeyFixture())
        self.useFixture(fixtures.NestedTempfile())
        self.temp_dir = self.useFixture(fixtures.TempDir()).path
        self.reporoot = os.path.join(self.temp_dir, 'reporoot')
        self.notesdir = os.path.join(self.reporoot,
                                     'releasenotes',
                                     'notes',
                                     )
        self._git_setup()
        self._counter = itertools.count(1)
        self.get_note_num = lambda: next(self._counter)


class BasicTest(Base):

    def test_non_python_no_tags(self):
        filename = self._add_notes_file()
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'0.0.0': [filename]},
            results,
        )

    def test_python_no_tags(self):
        self._make_python_package()
        filename = self._add_notes_file()
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'0.0.0': [filename]},
            results,
        )

    def test_note_before_tag(self):
        filename = self._add_notes_file()
        self._add_other_file('not-a-release-note.txt')
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'1.0.0': [filename]},
            results,
        )

    def test_note_commit_tagged(self):
        filename = self._add_notes_file()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'1.0.0': [filename]},
            results,
        )

    def test_note_commit_after_tag(self):
        self._make_python_package()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        filename = self._add_notes_file()
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'1.0.0-1': [filename]},
            results,
        )

    def test_other_commit_after_tag(self):
        filename = self._add_notes_file()
        self._add_other_file('ignore-1.txt')
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        self._add_other_file('ignore-2.txt')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'1.0.0': [filename]},
            results,
        )

    def test_multiple_notes_after_tag(self):
        self._make_python_package()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        f1 = self._add_notes_file()
        f2 = self._add_notes_file()
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'1.0.0-2': [f1, f2]},
            results,
        )

    def test_multiple_notes_within_tag(self):
        self._make_python_package()
        f1 = self._add_notes_file(commit=False)
        f2 = self._add_notes_file()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'1.0.0': [f1, f2]},
            results,
        )

    def test_multiple_tags(self):
        self._make_python_package()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        f1 = self._add_notes_file()
        self._run_git('tag', '-s', '-m', 'first tag', '2.0.0')
        f2 = self._add_notes_file()
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'2.0.0': [f1],
             '2.0.0-1': [f2],
             },
            results,
        )

    def test_rename_file(self):
        self._make_python_package()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        f1 = self._add_notes_file('slug1')
        self._run_git('tag', '-s', '-m', 'first tag', '2.0.0')
        f2 = f1.replace('slug1', 'slug2')
        self._run_git('mv', f1, f2)
        self._git_commit('rename note file')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'2.0.0': [f2],
             },
            results,
        )

    def test_rename_file_sort_earlier(self):
        self._make_python_package()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        f1 = self._add_notes_file('slug1')
        self._run_git('tag', '-s', '-m', 'first tag', '2.0.0')
        f2 = f1.replace('slug1', 'slug0')
        self._run_git('mv', f1, f2)
        self._git_commit('rename note file')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'2.0.0': [f2],
             },
            results,
        )

    def test_edit_file(self):
        self._make_python_package()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        f1 = self._add_notes_file()
        self._run_git('tag', '-s', '-m', 'first tag', '2.0.0')
        with open(os.path.join(self.reporoot, f1), 'w') as f:
            f.write('---\npreamble: new contents for file')
        self._git_commit('edit note file')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'2.0.0': [f1],
             },
            results,
        )

    def test_legacy_file(self):
        self._make_python_package()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        f1 = self._add_notes_file('slug1', legacy=True)
        self._run_git('tag', '-s', '-m', 'first tag', '2.0.0')
        f2 = f1.replace('slug1', 'slug2')
        self._run_git('mv', f1, f2)
        self._git_commit('rename note file')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'2.0.0': [f2],
             },
            results,
        )

    def test_rename_legacy_file_to_new(self):
        self._make_python_package()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        f1 = self._add_notes_file('slug1', legacy=True)
        self._run_git('tag', '-s', '-m', 'first tag', '2.0.0')
        # Rename the file with the new convention of placing the UUID
        # after the slug instead of before.
        f2 = f1.replace('0000000000000001-slug1',
                        'slug1-0000000000000001')
        self._run_git('mv', f1, f2)
        self._git_commit('rename note file')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'2.0.0': [f2],
             },
            results,
        )

    def test_limit_by_earliest_version(self):
        self._make_python_package()
        self._add_notes_file()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        f2 = self._add_notes_file()
        self._run_git('tag', '-s', '-m', 'middle tag', '2.0.0')
        f3 = self._add_notes_file()
        self._run_git('tag', '-s', '-m', 'last tag', '3.0.0')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
            earliest_version='2.0.0',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'2.0.0': [f2],
             '3.0.0': [f3],
             },
            results,
        )

    def test_delete_file(self):
        self._make_python_package()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        f1 = self._add_notes_file('slug1')
        f2 = self._add_notes_file('slug2')
        self._run_git('rm', f1)
        self._git_commit('remove note file')
        self._run_git('tag', '-s', '-m', 'first tag', '2.0.0')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'2.0.0': [f2],
             },
            results,
        )

    def test_rename_then_delete_file(self):
        self._make_python_package()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        f1 = self._add_notes_file('slug1')
        f2 = f1.replace('slug1', 'slug2')
        self._run_git('mv', f1, f2)
        self._git_commit('rename note file')
        self._run_git('rm', f2)
        self._git_commit('remove note file')
        f3 = self._add_notes_file('slug3')
        self._run_git('tag', '-s', '-m', 'first tag', '2.0.0')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'2.0.0': [f3],
             },
            results,
        )


class PreReleaseTest(Base):

    def test_alpha(self):
        self._make_python_package()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0.0a1')
        f1 = self._add_notes_file('slug1')
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0.0a2')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'1.0.0.0a2': [f1],
             },
            results,
        )

    def test_beta(self):
        self._make_python_package()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0.0b1')
        f1 = self._add_notes_file('slug1')
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0.0b2')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'1.0.0.0b2': [f1],
             },
            results,
        )

    def test_release_candidate(self):
        self._make_python_package()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0.0rc1')
        f1 = self._add_notes_file('slug1')
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0.0rc2')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'1.0.0.0rc2': [f1],
             },
            results,
        )

    def test_collapse(self):
        files = []
        self._make_python_package()
        files.append(self._add_notes_file('slug1'))
        self._run_git('tag', '-s', '-m', 'alpha tag', '1.0.0.0a1')
        files.append(self._add_notes_file('slug2'))
        self._run_git('tag', '-s', '-m', 'beta tag', '1.0.0.0b1')
        files.append(self._add_notes_file('slug3'))
        self._run_git('tag', '-s', '-m', 'release candidate tag', '1.0.0.0rc1')
        files.append(self._add_notes_file('slug4'))
        self._run_git('tag', '-s', '-m', 'full release tag', '1.0.0')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
            collapse_pre_releases=True,
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'1.0.0': files,
             },
            results,
        )

    def test_collapse_without_full_release(self):
        self._make_python_package()
        f1 = self._add_notes_file('slug1')
        self._run_git('tag', '-s', '-m', 'alpha tag', '1.0.0.0a1')
        f2 = self._add_notes_file('slug2')
        self._run_git('tag', '-s', '-m', 'beta tag', '1.0.0.0b1')
        f3 = self._add_notes_file('slug3')
        self._run_git('tag', '-s', '-m', 'release candidate tag', '1.0.0.0rc1')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
            collapse_pre_releases=True,
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'1.0.0.0a1': [f1],
             '1.0.0.0b1': [f2],
             '1.0.0.0rc1': [f3],
             },
            results,
        )

    def test_collapse_without_notes(self):
        self._make_python_package()
        self._run_git('tag', '-s', '-m', 'earlier tag', '0.1.0')
        f1 = self._add_notes_file('slug1')
        self._run_git('tag', '-s', '-m', 'alpha tag', '1.0.0.0a1')
        f2 = self._add_notes_file('slug2')
        self._run_git('tag', '-s', '-m', 'beta tag', '1.0.0.0b1')
        f3 = self._add_notes_file('slug3')
        self._run_git('tag', '-s', '-m', 'release candidate tag', '1.0.0.0rc1')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
            collapse_pre_releases=True,
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'1.0.0.0a1': [f1],
             '1.0.0.0b1': [f2],
             '1.0.0.0rc1': [f3],
             },
            results,
        )


class MergeCommitTest(Base):

    def test_1(self):
        # Create changes on master and in the branch
        # in order so the history is "normal"
        n1 = self._add_notes_file()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        self._run_git('checkout', '-b', 'test_merge_commit')
        n2 = self._add_notes_file()
        self._run_git('checkout', 'master')
        self._add_other_file('ignore-1.txt')
        self._run_git('merge', '--no-ff', 'test_merge_commit')
        self._add_other_file('ignore-2.txt')
        self._run_git('tag', '-s', '-m', 'second tag', '2.0.0')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'1.0.0': [n1],
             '2.0.0': [n2]},
            results,
        )
        self.assertEqual(
            ['2.0.0', '1.0.0'],
            list(raw_results.keys()),
        )

    def test_2(self):
        # Create changes on the branch before the tag into which it is
        # actually merged.
        self._add_other_file('ignore-0.txt')
        self._run_git('checkout', '-b', 'test_merge_commit')
        n1 = self._add_notes_file()
        self._run_git('checkout', 'master')
        n2 = self._add_notes_file()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        self._add_other_file('ignore-1.txt')
        self._run_git('merge', '--no-ff', 'test_merge_commit')
        self._add_other_file('ignore-2.txt')
        self._run_git('tag', '-s', '-m', 'second tag', '2.0.0')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'1.0.0': [n2],
             '2.0.0': [n1]},
            results,
        )
        self.assertEqual(
            ['2.0.0', '1.0.0'],
            list(raw_results.keys()),
        )

    def test_3(self):
        # Create changes on the branch before the tag into which it is
        # actually merged, with another tag in between the time of the
        # commit and the time of the merge. This should reflect the
        # order of events described in bug #1522153.
        self._add_other_file('ignore-0.txt')
        self._run_git('checkout', '-b', 'test_merge_commit')
        n1 = self._add_notes_file()
        self._run_git('checkout', 'master')
        n2 = self._add_notes_file()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        self._add_other_file('ignore-1.txt')
        self._run_git('tag', '-s', '-m', 'second tag', '1.1.0')
        self._run_git('merge', '--no-ff', 'test_merge_commit')
        self._add_other_file('ignore-2.txt')
        self._run_git('tag', '-s', '-m', 'third tag', '2.0.0')
        self._add_other_file('ignore-3.txt')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        # Since the 1.1.0 tag has no notes files, it does not appear
        # in the output. It's only there to trigger the bug as it was
        # originally reported.
        self.assertEqual(
            {'1.0.0': [n2],
             '2.0.0': [n1]},
            results,
        )
        self.assertEqual(
            ['2.0.0', '1.0.0'],
            list(raw_results.keys()),
        )

    def test_4(self):
        # Create changes on the branch before the tag into which it is
        # actually merged, with another tag in between the time of the
        # commit and the time of the merge. This should reflect the
        # order of events described in bug #1522153.
        self._add_other_file('ignore-0.txt')
        self._run_git('checkout', '-b', 'test_merge_commit')
        n1 = self._add_notes_file()
        self._run_git('checkout', 'master')
        n2 = self._add_notes_file()
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        self._add_other_file('ignore-1.txt')
        n3 = self._add_notes_file()
        self._run_git('tag', '-s', '-m', 'second tag', '1.1.0')
        self._run_git('merge', '--no-ff', 'test_merge_commit')
        self._add_other_file('ignore-2.txt')
        self._run_git('tag', '-s', '-m', 'third tag', '2.0.0')
        self._add_other_file('ignore-3.txt')
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {'1.0.0': [n2],
             '1.1.0': [n3],
             '2.0.0': [n1]},
            results,
        )
        self.assertEqual(
            ['2.0.0', '1.1.0', '1.0.0'],
            list(raw_results.keys()),
        )


class UniqueIdTest(Base):

    def test_legacy(self):
        uid = scanner._get_unique_id(
            'releasenotes/notes/0000000000000001-slug1.yaml'
        )
        self.assertEqual('0000000000000001', uid)

    def test_modern(self):
        uid = scanner._get_unique_id(
            'releasenotes/notes/slug1-0000000000000001.yaml'
        )
        self.assertEqual('0000000000000001', uid)


class BranchTest(Base):

    def setUp(self):
        super(BranchTest, self).setUp()
        self._make_python_package()
        self.f1 = self._add_notes_file('slug1')
        self._run_git('tag', '-s', '-m', 'first tag', '1.0.0')
        self.f2 = self._add_notes_file('slug2')
        self._run_git('tag', '-s', '-m', 'first tag', '2.0.0')
        self._add_notes_file('slug3')
        self._run_git('tag', '-s', '-m', 'first tag', '3.0.0')

    def test_files_current_branch(self):
        self._run_git('checkout', '2.0.0')
        self._run_git('checkout', '-b', 'stable/2')
        f21 = self._add_notes_file('slug21')
        log_text = self._run_git('log')
        self.addDetail('git log', text_content(log_text))
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {
                '1.0.0': [self.f1],
                '2.0.0': [self.f2],
                '2.0.0-1': [f21],
            },
            results,
        )

    def test_files_stable_from_master(self):
        self._run_git('checkout', '2.0.0')
        self._run_git('checkout', '-b', 'stable/2')
        f21 = self._add_notes_file('slug21')
        self._run_git('checkout', 'master')
        log_text = self._run_git('log', '--pretty=%x00%H %d', '--name-only',
                                 'stable/2')
        self.addDetail('git log', text_content(log_text))
        raw_results = scanner.get_notes_by_version(
            self.reporoot,
            'releasenotes/notes',
            'stable/2',
        )
        results = {
            k: [f for (f, n) in v]
            for (k, v) in raw_results.items()
        }
        self.assertEqual(
            {
                '1.0.0': [self.f1],
                '2.0.0': [self.f2],
                '2.0.0-1': [f21],
            },
            results,
        )


class GetTagsParseTest(base.TestCase):

    EXPECTED = [
        '2.0.0',
        '1.8.1',
        '1.8.0',
        '1.7.1',
        '1.7.0',
        '1.6.0',
        '1.5.0',
        '1.4.0',
        '1.3.0',
        '1.2.0',
        '1.1.0',
        '1.0.0',
        '0.11.2',
        '0.11.1',
        '0.11.0',
        '0.10.1',
        '0.10.0',
        '0.9.0',
        '0.8.0',
        '0.7.1',
        '0.7.0',
        '0.6.0',
        '0.5.1',
        '0.5.0',
        '0.4.2',
        '0.4.1',
        '0.4.0',
        '0.3.2',
        '0.3.1',
        '0.3.0',
        '0.2.5',
        '0.2.4',
        '0.2.3',
        '0.2.2',
        '0.2.1',
        '0.2.0',
        '0.1.3',
        '0.1.2',
        '0.1.1',
        '0.1.0',
    ]

    def test_keystoneclient_ubuntu_1_9_1(self):
        # git 1.9.1 as it produces output on ubuntu for python-keystoneclient
        # git log --simplify-by-decoration --pretty="%d"
        tag_list_output = textwrap.dedent("""
         (HEAD, origin/master, origin/HEAD, gerrit/master, master)
         (apu/master)
         (tag: 2.0.0)
         (tag: 1.8.1)
         (tag: 1.8.0)
         (tag: 1.7.1)
         (tag: 1.7.0)
         (tag: 1.6.0)
         (tag: 1.5.0)
         (tag: 1.4.0)
         (uncap-requirements)
         (tag: 1.3.0)
         (tag: 1.2.0)
         (tag: 1.1.0)
         (tag: 1.0.0)
         (tag: 0.11.2)
         (tag: 0.11.1)
         (tag: 0.11.0)
         (tag: 0.10.1)
         (tag: 0.10.0)
         (tag: 0.9.0)
         (tag: 0.8.0)
         (tag: 0.7.1)
         (tag: 0.7.0)
         (tag: 0.6.0)
         (tag: 0.5.1)
         (tag: 0.5.0)
         (tag: 0.4.2)
         (tag: 0.4.1)
         (tag: 0.4.0)
         (tag: 0.3.2)
         (tag: 0.3.1)
         (tag: 0.3.0)
         (tag: 0.2.5)
         (tag: 0.2.4)
         (tag: 0.2.3)
         (tag: 0.2.2)
         (tag: 0.2.1)
         (tag: 0.2.0)

         (origin/feature/keystone-v3, gerrit/feature/keystone-v3)
         (tag: 0.1.3)
         (tag: 0.1.2)
         (tag: 0.1.1)
         (tag: 0.1.0)
         (tag: folsom-1)
         (tag: essex-rc1)
         (tag: essex-4)
         (tag: essex-3)
        """)
        with mock.patch('reno.utils.check_output') as co:
            co.return_value = tag_list_output
            actual = scanner._get_version_tags_on_branch('reporoot',
                                                         branch=None)
        self.assertEqual(self.EXPECTED, actual)

    def test_keystoneclient_rhel_1_7_1(self):
        # git 1.7.1 as it produces output on RHEL 6 for python-keystoneclient
        # git log --simplify-by-decoration --pretty="%d"
        tag_list_output = textwrap.dedent("""
         (HEAD, origin/master, origin/HEAD, master)
         (tag: 2.0.0)
         (tag: 1.8.1)
         (tag: 1.8.0)
         (tag: 1.7.1)
         (tag: 1.7.0)
         (tag: 1.6.0)
         (tag: 1.5.0)
         (tag: 1.4.0)
         (tag: 1.3.0)
         (tag: 1.2.0)
         (tag: 1.1.0)
         (tag: 1.0.0)
         (tag: 0.11.2)
         (tag: 0.11.1)
         (tag: 0.11.0)
         (tag: 0.10.1)
         (tag: 0.10.0)
         (tag: 0.9.0)
         (tag: 0.8.0)
         (tag: 0.7.1)
         (tag: 0.7.0)
         (tag: 0.6.0)
         (tag: 0.5.1)
         (tag: 0.5.0)
         (tag: 0.4.2)
         (tag: 0.4.1)
         (tag: 0.4.0)
         (tag: 0.3.2)
         (tag: 0.3.1)
         (tag: 0.3.0)
         (tag: 0.2.5)
         (tag: 0.2.4)
         (tag: 0.2.3)
         (tag: 0.2.2)
         (tag: 0.2.1)
         (tag: 0.2.0)
         (tag: 0.1.3)
         (0.1.2)
         (tag: 0.1.1)
         (0.1.0)
         (tag: folsom-1)
         (tag: essex-rc1)
         (essex-4)
         (essex-3)
        """)
        with mock.patch('reno.utils.check_output') as co:
            co.return_value = tag_list_output
            actual = scanner._get_version_tags_on_branch('reporoot',
                                                         branch=None)
        self.assertEqual(self.EXPECTED, actual)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os.path

from reno import scanner

import six
import yaml

LOG = logging.getLogger(__name__)


def get_cache_filename(reporoot, notesdir):
    return os.path.join(reporoot, notesdir, 'reno.cache')


class Loader(object):
    "Load the release notes for a given repository."

    def __init__(self, reporoot, notesdir, branch=None,
                 collapse_pre_releases=True,
                 earliest_version=None,
                 ignore_cache=False):
        """Initialize a Loader.

        The versions are presented in reverse chronological order.

        Notes files are associated with the earliest version for which
        they were available, regardless of whether they changed later.

        :param reporoot: Path to the root of the git repository.
        :type reporoot: str
        :param notesdir: The directory under *reporoot* with the release notes.
        :type notesdir: str
        :param branch: The name of the branch to scan. Defaults to current.
        :type branch: str
        :param collapse_pre_releases: When true, merge pre-release versions
            into the final release, if it is present.
        :type collapse_pre_releases: bool
        :param earliest_version: The oldest version to include.
        :type earliest_version: str
        :param ignore_cache: Do not load a cache file if it is present.
        :type ignore_cache: bool

        """
        self._reporoot = reporoot
        self._notesdir = notesdir
        self._branch = branch
        self._collapse_pre_releases = collapse_pre_releases
        self._earliest_version = earliest_version
        self._ignore_cache = ignore_cache

        self._cache = None
        self._scanner_output = None
        self._cache_filename = get_cache_filename(reporoot, notesdir)

        self._load_data()

    def _load_data(self):
        cache_file_exists = os.path.exists(self._cache_filename)

        if self._ignore_cache and cache_file_exists:
            LOG.debug('ignoring cache file %s', self._cache_filename)

        if (not self._ignore_cache) and cache_file_exists:
            with open(self._cache_filename, 'r') as f:
                self._cache = yaml.safe_load(f.read())
                # Save the cached scanner output to the same attribute
                # it would be in if we had loaded it "live". This
                # simplifies some of the logic in the other methods.
                self._scanner_output = {
                    n['version']: n['files']
                    for n in self._cache['notes']
                }
        else:
            self._scanner_output = scanner.get_notes_by_version(
                reporoot=self._reporoot,
                notesdir=self._notesdir,
                branch=self._branch,
                collapse_pre_releases=self._collapse_pre_releases,
                earliest_version=self._earliest_version,
            )

    @property
    def versions(self):
        "A list of all of the versions found."
        return list(self._scanner_output.keys())

    def __getitem__(self, version):
        "Return data about the files that should go into a given version."
        return self._scanner_output[version]

    def parse_note_file(self, filename, sha):
        """Return the data structure encoded in the note file.

        Emit warnings for content that does not look valid in some
        way, but return it anway for backwards-compatibility.

        """
        if self._cache:
            content = self._cache['file-contents'][filename]
        else:
            body = scanner.get_file_at_commit(self._reporoot, filename, sha)
            content = yaml.safe_load(body)

        for section_name, section_content in content.items():
            if section_name == 'prelude':
                if not isinstance(section_content, six.string_types):
                    LOG.warning(
                        ('The prelude section of %s '
                         'does not parse as a single string. '
                         'Is the YAML input escaped properly?') %
                        filename,
                    )
            else:
                if not isinstance(section_content, list):
                    LOG.warning(
                        ('The %s section of %s '
                         'does not parse as a list of strings. '
                         'Is the YAML input escaped properly?') % (
                             section_name, filename),
                    )
                else:
                    for item in section_content:
                        if not isinstance(item, six.string_types):
                            LOG.warning(
                                ('The item %r in the %s section of %s '
                                 'parses as a %s instead of a string. '
                                 'Is the YAML input escaped properly?'
                                 ) % (item, section_name,
                                      filename, type(item)),
                            )

        return content



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from reno import create
from reno.tests import base

import fixtures
import mock


class TestPickFileName(base.TestCase):

    @mock.patch('os.path.exists')
    def test_not_random_enough(self, exists):
        exists.return_value = True
        self.assertRaises(
            ValueError,
            create._pick_note_file_name,
            'somepath',
            'someslug',
        )

    @mock.patch('os.path.exists')
    def test_random_enough(self, exists):
        exists.return_value = False
        result = create._pick_note_file_name('somepath', 'someslug')
        self.assertIn('somepath', result)
        self.assertIn('someslug', result)


class TestCreate(base.TestCase):

    def setUp(self):
        super(TestCreate, self).setUp()
        self.tmpdir = self.useFixture(fixtures.TempDir()).path

    def test_create_from_template(self):
        filename = create._pick_note_file_name(self.tmpdir, 'theslug')
        create._make_note_file(filename)
        with open(filename, 'r') as f:
            body = f.read()
        self.assertEqual(create._TEMPLATE, body)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os.path

from reno import defaults
from reno import formatter
from reno import loader

from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst import directives
from docutils.statemachine import ViewList
from sphinx.util.nodes import nested_parse_with_titles


class ReleaseNotesDirective(rst.Directive):

    has_content = True

    option_spec = {
        'branch': directives.unchanged,
        'reporoot': directives.unchanged,
        'relnotessubdir': directives.unchanged,
        'notesdir': directives.unchanged,
        'version': directives.unchanged,
        'collapse-pre-releases': directives.flag,
        'earliest-version': directives.unchanged,
    }

    def run(self):
        env = self.state.document.settings.env
        app = env.app

        def info(msg):
            app.info('[reno] %s' % (msg,))

        title = ' '.join(self.content)
        branch = self.options.get('branch')
        reporoot_opt = self.options.get('reporoot', '.')
        reporoot = os.path.abspath(reporoot_opt)
        relnotessubdir = self.options.get('relnotessubdir',
                                          defaults.RELEASE_NOTES_SUBDIR)
        notessubdir = self.options.get('notesdir', defaults.NOTES_SUBDIR)
        version_opt = self.options.get('version')
        # FIXME(dhellmann): Force this flag True for now and figure
        # out how Sphinx passes a "false" flag later.
        collapse = True  # 'collapse-pre-releases' in self.options
        earliest_version = self.options.get('earliest-version')

        notesdir = os.path.join(relnotessubdir, notessubdir)
        info('scanning %s for %s release notes' %
             (os.path.join(reporoot, notesdir), branch or 'current branch'))

        ldr = loader.Loader(
            reporoot=reporoot,
            notesdir=notesdir,
            branch=branch,
            collapse_pre_releases=collapse,
            earliest_version=earliest_version,
        )
        if version_opt is not None:
            versions = [
                v.strip()
                for v in version_opt.split(',')
            ]
        else:
            versions = ldr.versions
        text = formatter.format_report(
            ldr,
            versions,
            title=title,
        )
        source_name = '<' + __name__ + '>'
        result = ViewList()
        for line in text.splitlines():
            result.append(line, source_name)

        node = nodes.section()
        node.document = self.state.document
        nested_parse_with_titles(self.state, result, node)
        return node.children


def setup(app):
    app.add_directive('release-notes', ReleaseNotesDirective)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function


_SECTION_ORDER = [
    ('features', 'New Features'),
    ('issues', 'Known Issues'),
    ('upgrade', 'Upgrade Notes'),
    ('deprecations', 'Deprecation Notes'),
    ('critical', 'Critical Issues'),
    ('security', 'Security Issues'),
    ('fixes', 'Bug Fixes'),
    ('other', 'Other Notes'),
]


def _indent_for_list(text, prefix='  '):
    """Indent some text to make it work as a list entry.

    Indent all lines except the first with the prefix.
    """
    lines = text.splitlines()
    return '\n'.join([lines[0]] + [
        prefix + l
        for l in lines[1:]
    ]) + '\n'


def format_report(loader, versions_to_include, title=None):
    report = []
    if title:
        report.append('=' * len(title))
        report.append(title)
        report.append('=' * len(title))
        report.append('')

    # Read all of the notes files.
    file_contents = {}
    for version in versions_to_include:
        for filename, sha in loader[version]:
            body = loader.parse_note_file(filename, sha)
            file_contents[filename] = body

    for version in versions_to_include:
        report.append(version)
        report.append('=' * len(version))
        report.append('')

        # Add the preludes.
        notefiles = loader[version]
        for n, sha in notefiles:
            if 'prelude' in file_contents[n]:
                report.append(file_contents[n]['prelude'])
                report.append('')

        for section_name, section_title in _SECTION_ORDER:
            notes = [
                n
                for fn, sha in notefiles
                if file_contents[fn].get(section_name)
                for n in file_contents[fn].get(section_name, [])
            ]
            if notes:
                report.append(section_title)
                report.append('-' * len(section_title))
                report.append('')
                for n in notes:
                    report.append('- %s' % _indent_for_list(n))
                report.append('')

    return '\n'.join(report)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function

import logging

from reno import loader
from reno import utils

LOG = logging.getLogger(__name__)


def list_cmd(args):
    "List notes files based on query arguments"
    LOG.debug('starting list')
    reporoot = args.reporoot.rstrip('/') + '/'
    notesdir = utils.get_notes_dir(args)
    collapse = args.collapse_pre_releases
    ldr = loader.Loader(
        reporoot=reporoot,
        notesdir=notesdir,
        branch=args.branch,
        collapse_pre_releases=collapse,
        earliest_version=args.earliest_version,
    )
    if args.version:
        versions = args.version
    else:
        versions = ldr.versions
    for version in versions:
        notefiles = ldr[version]
        print(version)
        for n, sha in notefiles:
            if n.startswith(reporoot):
                n = n[len(reporoot):]
            print('\t%s (%s)' % (n, sha))
    return
